local modName = "Party & Raid"
local parent = HudMap
local L = LibStub("AceLocale-3.0"):GetLocale("HudMap")
local mod = HudMap:NewModule(modName, "AceEvent-3.0")

--[[ Default upvals 
     This has a slight performance benefit, but upvalling these also makes it easier to spot leaked globals. ]]--
local _G = _G.getfenv(0)
local wipe, type, pairs, tinsert, tremove, tonumber = _G.wipe, _G.type, _G.pairs, _G.tinsert, _G.tremove, _G.tonumber
local math, math_abs, math_pow, math_sqrt, math_sin, math_cos, math_atan2 = _G.math, _G.math.abs, _G.math.pow, _G.math.sqrt, _G.math.sin, _G.math.cos, _G.math.atan2
local error, rawset, rawget, print = _G.error, _G.rawset, _G.rawget, _G.print
local tonumber, tostring = _G.tonumber, _G.tostring
local getmetatable, setmetatable, pairs, ipairs, select, unpack = _G.getmetatable, _G.setmetatable, _G.pairs, _G.ipairs, _G.select, _G.unpack
--[[ -------------- ]]--

local db
local group = parent.group

local frame = CreateFrame("Frame")
local partyMembers = {}
local targetArrows = {}

local options = {
	type = "group",
	name = L["Party & Raid"],
	args = {
		enable = {
			type = "toggle",
			name = L["Enable Party & Raid Dots"],
			order = 100,
			get = function()
				return db.enable
			end,
			set = function(info, v)
				db.enable = v
				local r = v and mod:Enable() or mod:Disable()
			end
		},
		showSpells = {
			type = "toggle",
			name = L["Show Spellcasts"],
			order = 110,
			get = function()
				return db.showSpells
			end,
			set = function(info, v)
				db.showSpells = v
			end
		},
		showSpellTargets = {
			type = "toggle",
			name = L["Show Spell Targets"],
			order = 111,
			get = function()
				return db.showSpellTargets
			end,
			set = function(info, v)
				db.showSpellTargets = v
			end
		},
		showTargets = {
			type = "toggle",
			name = L["Show Targets"],
			order = 109,
			get = function()
				return db.showTargets
			end,
			set = function(info, v)
				db.showTargets = v
			end
		},
		dotSize = {
			type = "range",
			name = L["Dot Size"],
			order = 120,
			min = 4,
			max = 50,
			step = 1,
			bigStep = 1,
			get = function()
				return db.iconSize
			end,
			set = function(info, v)
				db.iconSize = v
				for k, dot in pairs(partyMembers) do
					dot:SetSize(v .. "px")
				end
			end
		}
	}
}

local defaults = {
	profile = {
		enable = true,
		showSpells = false,
		showSpellTargets = false,
		showTargets = false,
		iconSize = 13
	}
}

local classLookup = setmetatable({}, {__index = function(t, k)
	local _, c = UnitClass(k)
	if c then
		rawset(t, k, c)
		return c
	end
	return nil
end})

local SN = parent.SN

local spellMap = {
	-- Paladin
	[48825] = "healer",		-- Holy Shock
	[53562] = "healer",		-- Beacon of Light
	[48952] = "tank",			-- Holy Shield
	[35395] = "dps",			-- Crusader Strike
	
	-- Priest
	[53077] = "healer",		-- Penance
	[48089] = "healer",		-- Circle of Healing
	[15473] = "dps",			-- Shadowform
	[65490] = "dps",			-- Vampiric Touch
	
	-- Druid
	[53251] = "healer",		-- Wild Growth
	[33891] = "healer",		-- Tree of Life
	[6807]  = "tank",			-- Maul
	[9634]  = "tank",			-- Bear Form
	[48562] = "tank",			-- Swipe
	[48562] = "tank",			-- Swipe
	[48572] = "dps",			-- Shred
	[48468] = "dps",			-- Insect Swarm
	[24858] = "dps",			-- Moonkin Form
	[768]   = "dps", 			-- Cat Form
	
	-- Death Knight
	[48263] = "tank",			-- Frost presence
	[48266] = "dps",			-- Blood presence
	[48265] = "dps",			-- Unholy presence
	
	-- Warrior
	[71]    = "tank",			-- Defensive Stance
	[57823] = "tank",			-- Revenge
	[47486] = "dps",			-- Mortal Strike
	[23881] = "dps",			-- Bloodthirst
	
	-- Shaman
	[61301] = "healer",   -- Riptide
	[49284] = "healer",   -- Earth Shield
	[57722] = "dps",   		-- Totem of Wrath
	[17364] = "dps",   		-- Stormstrike
	[59159] = "dps"
}

local bestGuess = {
	DEATHKNIGHT = function(unit)
		if UnitAura(unit, SN[48263]) then return "tank" end
		if UnitAura(unit, SN[48266]) then return "dps" end
		if UnitAura(unit, SN[48265]) then return "dps" end
	end,
	DRUID = function(unit)
		if UnitAura(unit, SN[9634])  then return "tank" end
		if UnitAura(unit, SN[768])   then return "dps" end
		if UnitAura(unit, SN[24858]) then return "dps" end
		if UnitAura(unit, SN[33891]) then return "healer" end
	end,
	MAGE = "dps",
	HUNTER = "dps",	
	PALADIN = function(unit)
		if UnitLevel(unit) == 80 and UnitManaMax(unit) > 15000 then return "healer" end
		if UnitAura(unit, SN[25780]) then return "tank" end
		return "dps"
	end,
	PRIEST = function(unit)
		if UnitAura(unit, SN[15473]) then return "dps" end
		return "healer"
	end,
	ROGUE = "dps",
	SHAMAN = "dps",
	WARLOCK = "dps",
	WARRIOR = function(unit)
		if UnitLevel(unit) == 80 and UnitHealthMax(unit) > 40000 then return "tank" end
		return "dps"
	end
}

local spellNameMap = {}
for k, v in pairs(spellMap) do
	spellNameMap[SN[k]] = v
end

local function UnitIsMappable(unit, allowSelf)
	local x, y = GetPlayerMapPosition(unit)
	if (not allowSelf and UnitIsUnit("player", unit)) or
		 (x == 0 and y == 0) or
		 not UnitIsConnected(unit) or
		 UnitIsConnected(unit) == 0
		 then return false end
	return true
end

function mod:UpdatePartyUnit(unit)
	local n = UnitName(unit)
	if not n then return end
	if not UnitIsMappable(unit) and partyMembers[n] then
		partyMembers[n] = partyMembers[n]:Free(true)
	elseif UnitIsMappable(unit) and not partyMembers[n] then
		local cl = RAID_CLASS_COLORS[classLookup[n]]
		if cl then
			local _, uc = UnitClass(unit)
			local r, g , b, a = cl.r, cl.g, cl.b, 1
			r = r + ((1-r) / 5)
			g = g + ((1-g) / 5)
			b = b + ((1-b) / 5)
			local guess = bestGuess[uc]
			if type(guess) == "function" then guess = guess(unit) end
			local c = parent:PlaceRangeMarkerOnPartyMember(guess or "party", unit, db.iconSize .. "px", nil, r, g, b, a, "BLEND")
			c.RegisterCallback(self, "Free", "FreeDot")
			partyMembers[n] = c
		end
	end
end

function mod:FreeDot(dot)
	if dot.follow then
		partyMembers[dot.follow] = nil
	end
end

function mod:UpdateParty()
	for index, unit in group() do
		self:UpdatePartyUnit(unit)
	end
	for name, dot in pairs(partyMembers) do
		if not UnitIsMappable(name) or partyMembers[name].freed then
			partyMembers[name] = dot:Free()
		end
	end
end

local throttle = 10
local counter = 0
local function update(self, t)
	counter = counter + t
	if counter > throttle then
		counter = counter - throttle
		mod:UpdateParty()
	end
end

function mod:OnInitialize()
	self.db = parent.db:RegisterNamespace(modName, defaults)
	db = self.db.profile
	parent:RegisterModuleOptions(modName, options, modName)
end

function mod:OnEnable()
	frame:SetScript("OnUpdate", update)
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self:RegisterEvent("UNIT_TARGET")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterEvent("RAID_ROSTER_UPDATE", "UpdateParty")
	self:RegisterEvent("PARTY_MEMBERS_CHANGED", "UpdateParty")
	self:UpdateParty()
end

function mod:OnDisable()
	frame:SetScript("OnUpdate", nil)
	for k, v in pairs(partyMembers) do
		partyMembers[k] = v:Free()
	end
end

local roles = {}
local playerName
local function validUnit(unit)
	return unit and (partyMembers[unit] or UnitIsUnit(unit, "player")) and UnitIsMappable(unit, true)
end

function mod:COMBAT_LOG_EVENT_UNFILTERED(ev, timestamp, event, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, spellID, spellName, ...)
	if spellNameMap[spellName] and partyMembers[sourceName] and roles[sourceName] ~= spellNameMap[spellName] then
		partyMembers[sourceName]:SetTexture(spellNameMap[spellName], "BLEND")
		roles[sourceName] = spellNameMap[spellName]
	end
	
	if event == "SPELL_CAST_START" or event == "SPELL_CAST_SUCCESS" or event == "SPELL_HEAL" then
		if db.showSpells and sourceName then
			local dot = partyMembers[sourceName] 
			if dot then
				local _, _, icon = GetSpellInfo(spellID)
				dot:SetTexture(icon)
				dot.texture:SetVertexColor(1,1,1,1)
			end
		end
		if db.showSpellTargets and sourceName and destName then
			if validUnit(destName) then
				local edge = parent:AddEdge(0.5, 1, 0.5, 0.7, 1, sourceName, destName)				
				if partyMembers[sourceName] then partyMembers[sourceName]:AttachEdge(edge) end
				if partyMembers[destName] then partyMembers[destName]:AttachEdge(edge) end
			end
		end
	end
end

function mod:PLAYER_TARGET_CHANGED()
	self:UNIT_TARGET(nil, "player")
end

function mod:UNIT_TARGET(event, unit)
	if not db.showTargets then return end
	local n = UnitName(unit)
	if validUnit(n) then
		if targetArrows[n] then targetArrows[n] = targetArrows[n]:Free() end
		
		local t = UnitName(unit .. "target")
		if t and validUnit(t) then
			local edge = parent:AddEdge(1, 1, 1, 0.4, nil, unit, t)
			targetArrows[n] = edge
			if partyMembers[n] then partyMembers[n]:AttachEdge(edge) end
			if partyMembers[t] then partyMembers[t]:AttachEdge(edge) end
		end		
	end
end